import numpy as np
from abc import ABC, abstractmethod
import optuna

rng = np.random.default_rng()

class Strategy(ABC):

    @abstractmethod
    def vector(self):
        pass

class ConstantThresholdStrategy(Strategy):

    def __init__(self, N, threshold):
        self.N = N
        self.threshold = threshold
        self._vector = np.empty(2*self.N, dtype=int)
        self._vector[:-1] = self.threshold
        self._vector[-1] = 0

    def vector(self):
        return self._vector


class LinearThresholdStrategy(Strategy):

    def __init__(self, N, a, b):
        self.N = N
        self.a = a
        self.b = b
        self._vector = np.empty(2 * self.N, dtype=int)
        self._vector = np.floor(self.a * (2*self.N - np.arange(2*self.N)) + self.b)
        self._vector[-1] = 0

    def vector(self):
        return self._vector



class GameSimulator:
    """
    docstring for GameSimulator
    """

    def __init__(self, N):
        self.N = N
        self.deck = np.array([-1] * self.N + [1] * self.N, dtype=np.int8)
        # self.deck = np.concatenate([-np.ones(self.N), np.ones(self.N)])

    def play(self, strategy):
        rng.shuffle(self.deck)
        scores = np.cumsum(self.deck)
        return scores[scores>=strategy.vector()][0]


    def monte_carlo(self, strategy, M=50000):
        mat = np.tile(self.deck, (M, 1))
        np.apply_along_axis(rng.shuffle, 1, mat)
        scores = np.cumsum(mat, axis=1)
        mask = scores >= strategy.vector()
        indices = np.argmax(mask, axis=1)
        return np.mean(scores[np.arange(M),indices])


if __name__ == "__main__":
    N = 16

    game = GameSimulator(N)

    for threshold in range(1, 8):
        s = ConstantThresholdStrategy(N, threshold)
        print(f"threshold = {threshold} --  expected score = {game.monte_carlo(s)}")

    s = LinearThresholdStrategy(N, 0.1, 2)
    print(f"(a,b) = {(s.a,s.b)} --  expected score = {game.monte_carlo(s)}")

    print(s.vector())


    def objective(trial):
        a = trial.suggest_uniform('a', 0., 2.)  # x is between 0 and pi
        b = trial.suggest_uniform('b', 0., 5.)  # y is between 0 and pi
        s = LinearThresholdStrategy(N, a, b)

        return -game.monte_carlo(s)


    study = optuna.create_study(direction='minimize')
    study.optimize(objective, n_trials=100)

    print(LinearThresholdStrategy(N, **study.best_params).vector())
